#!/usr/bin/env micropython
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    from typing import Optional, Tuple
except ImportError:
    pass
import math

import picographics
import hub75


class Graphics:
    def __init__(self,
                 display_type: picographics.DisplayType,
                 rotate: Optional[int] = 0,
                 panel_type: hub75.PanelType = hub75.PANEL_GENERIC,
                 stb_invert: bool = False,
                 color_order: hub75.ColorOrder
                 = hub75.COLOR_ORDER_RGB) -> None:
        self._driver = picographics.PicoGraphics(display_type)
        self.rotate = rotate

        width, height = self._driver.get_bounds()
        self.hub75 = hub75.Hub75(width,
                                 height,
                                 panel_type=panel_type,
                                 stb_invert=stb_invert,
                                 color_order=color_order)
        self.hub75.start()

    def create_pen(self, r: int, g: int, b: int) -> picographics.Pen:
        return self._driver.create_pen(r, g, b)

    def set_pen(self, pen: picographics.Pen) -> None:
        self._driver.set_pen(pen)

    def line(self, x1: int, y1: int, x2: int, y2: int) -> None:
        # While picographics has a line function, it doesn't include
        # the second point when drawing.

        if x1 == x2:
            for y in range(y1 if y1 < y2 else y2, (y2 if y1 < y2 else y1)+1):
                self.pixel(x1, y)
            return
        if y1 == y2:
            for x in range(x1 if x1 < x2 else x2, (x2 if x1 < x2 else x1)+1):
                self.pixel(x, y1)
            return

        # This is Bresenham's Algorithm
        x, y = x1, y1
        dx = abs(x2 - x1)
        dy = abs(y2 - y1)

        pixel = self.pixel
        if (dy / float(dx)) > 1:
            dx, dy = dy, dx
            x, y = y, x
            x1, y1, x2, y2 = y1, x1, y2, x2
            pixel = self.__pixel_reverse

        p = 2*dy - dx

        pixel(x, y)

        for _ in range(2, dx + 2):
            if p > 0:
                y = y + 1 if y < y2 else y - 1
                p = p + 2 * (dy - dx)
            else:
                p = p + 2 * dy

            x = x + 1 if x < x2 else x - 1

            pixel(x, y)

    def circle(self, cx: int, cy: int, radius: int) -> None:
        d = 3 - 2 * radius
        y = radius
        i = 0
        while i <= y:
            self.line(cx + i, cy + y, cx + i, cy - y)
            self.line(cx - i, cy + y, cx - i, cy - y)
            self.line(cx + y, cy + i, cx + y, cy - i)
            self.line(cx - y, cy - i, cx - y, cy + i)

            if d < 0:
                d = d + 4 * i + 6
            else:
                d = d + 4 * (i - y) + 10
                y = y - 1
            i = i + 1

    def pixel(self, x: int, y: int) -> None:
        if x >= 0 and x < 64 and y >= 0 and y < 64:
            self._driver.pixel(x, y)

    def __pixel_reverse(self, x: int, y: int) -> None:
        if x >= 0 and x < 64 and y >= 0 and y < 64:
            self._driver.pixel(y, x)

    def update(self) -> None:
        self.hub75.update(self._driver)

    def get_bounds(self) -> Tuple[int, int]:
        return self._driver.get_bounds()



#!/usr/bin/env micropython
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
from typing import List, Optional, Tuple

import pygame

from pen import Pen, RGB888Pen


class DisplayType:
    def __init__(self, width: int, height: int,) -> None:
        self.width = width
        self.height = height


DISPLAY_INTERSTATE75_64X64 = DisplayType(64, 64)


class PicoGraphics:
    def __init__(self, display_type: DisplayType):
        self.display_type = display_type
        self.pen = self.create_pen(0, 0, 0)

        self._buffer: List[List[Optional[Tuple[int, int, int]]]] = []
        for _ in range(self.display_type.height):
            self._buffer.append([None] * self.display_type.width)

    def create_pen(self, r: int, g: int, b: int) -> Pen:
        return RGB888Pen(r, g, b)

    def set_pen(self, pen: Pen) -> None:
        self.pen = pen

    def line(self, x1: int, y1: int, x2: int, y2: int) -> None:
        # This doesn't properly replicate picographs.
        # It seems to sort the two points, and then doesn't
        # include the second point when drawing the line.
        llen = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        dx = (x2 - x1) / llen
        dy = (y2 - y1) / llen

        x: float = x1
        y: float = y1
        last_coord = (x1, y1)
        self.pixel(x1, y1)
        for _ in range(math.floor(llen) - 1):
            x, y = x + dx, y + dy
            px, py = math.floor(x), math.floor(y)
            if last_coord != (px, py):
                last_coord = (px, py)
                self.pixel(px, py)

    def pixel(self, x: int, y: int) -> None:
        self._buffer[y][x] = self.pen.as_tuple()

    def get_bounds(self) -> Tuple[int, int]:
        return (self.display_type.width, self.display_type.height)



#!/usr/bin/env micropython
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    from typing import Callable, Optional
except ImportError:  # pragma: no cover
    pass

import hub75
import picographics
from pimoroni import RGBLED, Button
from pimoroni_i2c import PimoroniI2C
import network
import ntptime
import machine
import time
import sys


from .datetime import DateTime
from .graphics import Graphics


# Index Constants
SWITCH_A = 0
SWITCH_B = 1
SWITCH_BOOT = 1


class BaseI75:
    I2C_SDA_PIN = 20
    I2C_SCL_PIN = 21
    SWITCH_PINS = (14, 23)
    SWITCH_PINS_W = (14, 15)
    LED_R_PIN = 16
    LED_G_PIN = 17
    LED_B_PIN = 18

    NUM_SWITCHES = 2

    """This is the base class for the Interstate75 library. All
       functions are overridden by either an emulated version, which
       replicates the functionality, or a native version if we're
       running on real Interstate75 hardware."""
    def __init__(self,
                 display_type: picographics.DisplayType,
                 stb_invert=False,
                 rotate: int = 0,
                 wifi_ssid: Optional[str] = None,
                 wifi_password: Optional[str] = None) -> None:
        self.wifi_ssid = wifi_ssid
        self.wifi_password = wifi_password

        self.wlan: Optional[network.WLAN] = None

        self.interstate75w = not hasattr(sys.implementation, "_machine") \
            or "Pico W" in sys.implementation._machine

        self.display = Graphics(display_type,
                                rotate,
                                hub75.PANEL_GENERIC,
                                stb_invert,
                                hub75.COLOR_ORDER_RGB)

        if self.interstate75w:
            self._switch_pins = self.SWITCH_PINS_W
        else:
            self._switch_pins = self.SWITCH_PINS

        # Set up the user switches
        self.__switches = []
        for i in range(self.NUM_SWITCHES):
            self.__switches.append(Button(self._switch_pins[i]))

        self.__rgb = RGBLED(BaseI75.LED_R_PIN,
                            BaseI75.LED_G_PIN,
                            BaseI75.LED_B_PIN,
                            invert=True)

        # Set up the i2c for Qw/st and Breakout Garden
        self.i2c = PimoroniI2C(self.I2C_SDA_PIN, self.I2C_SCL_PIN, 100000)

        self.rtc: Optional[machine.RTC] = None

    def enable_wifi(self,
                    callback: Optional[Callable[[int], None]] = None
                    ) -> bool:
        """Enables wifi.

           To run on real Interstate75 hardware you must have a secrets files
           (see :doc:`secrets`)

           :param callback: A function which is called every 200ms with the
               current connection status.
        """
        if self.wifi_ssid is None or self.wifi_password is None:
            return False

        # Start connection
        self.wlan = network.WLAN(network.STA_IF)
        self.wlan.active(True)
        # Turn WiFi power saving off for some slow APs
        self.wlan.config(pm=0xa11140)
        self.wlan.connect(self.wifi_ssid, self.wifi_password)

        # Wait for connect success or failure
        max_wait = 100
        while max_wait > 0:
            if self.wlan.status() < 0 or self.wlan.status() >= 3:
                break
            max_wait -= 1
            if callback is not None:
                callback(self.wlan.status())
            time.sleep_ms(200)

        return self.wlan.isconnected()

    def disable_wifi(self) -> None:
        """
        Disables the wifi.

        Noop is wifi is not available, or hasn't been enabled.
        """
        if self.wifi_ssid is None or self.wifi_password is None\
                or self.wlan is None:
            return
        self.wlan.disconnect()
        self.wlan.active(False)

    def set_time(self) -> bool:  # pragma: no cover
        raise NotImplementedError()

    def now(self) -> DateTime:  # pragma: no cover
        raise NotImplementedError()

    def sleep_ms(self, delay: int) -> None:  # pragma: no cover
        raise NotImplementedError()

    def ticks_ms(self) -> int:  # pragma: no cover
        raise NotImplementedError()

    def ticks_diff(self, t1: int, t2: int) -> int:  # pragma: no cover
        raise NotImplementedError()

    @staticmethod
    def is_emulated() -> bool:  # pragma: no cover
        """
        Returns true if we are not running on real Interstate75 hardware.
        """
        raise NotImplementedError()



#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class NetworkType:
    pass


STA_IF = NetworkType()

STAT_IDLE = 0
STAT_CONNECTING = 1
STAT_WRONG_PASSWORD = -1
STAT_NO_AP_FOUND = -2
STAT_CONNECT_FAIL = -3
STAT_GOT_IP = 3


class WLAN:
    """
    This is an emulation of the built-in WLAN class. As we can't actually
    control the wifi, it just pretends to configure it when called.
    """
    def __init__(self, network_type: NetworkType) -> None:
        assert network_type is STA_IF, \
            "STA_IF is the only supported network type currently."

        self._wifi_active = False
        self._wifi_connected = False

    def active(self, active: bool) -> None:
        self._wifi_active = active

    def config(self, pm: int) -> None:
        pass

    def connect(self, ssid: str, password: str) -> None:
        assert self._wifi_active
        self._wifi_connected = True

    def disconnect(self) -> None:
        self._wifi_connected = False

    def status(self) -> int:
        if self._wifi_active and self._wifi_connected:
            return STAT_GOT_IP
        if self._wifi_active:
            return STAT_IDLE
        return STAT_CONNECT_FAIL

    def isconnected(self) -> bool:
        return self._wifi_connected



# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .basei75 import BaseI75


class Colour:
    def __init__(self, value: int) -> None:
        self._value = value

    @property
    def r(self) -> int:
        return (self._value >> 24) & 255

    @property
    def g(self) -> int:
        return (self._value >> 16) & 255

    @property
    def b(self) -> int:
        return (self._value >> 8) & 255

    @property
    def a(self) -> int:
        return self._value & 255

    def set_colour(self, i75: BaseI75):
        i75.display.set_pen(i75.display.create_pen(self.r, self.g, self.b))

    @staticmethod
    def fromint32(value) -> "Colour":
        return Colour(value)



#!/usr/bin/env python3
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
from typing import List, Optional, Tuple

import pygame

import picographics

LED_SIZE = 10


class ColorOrder:
    pass


class Hub75:
    def __init__(self,
                 width: int,
                 height: int,
                 panel_type: "PanelType",
                 stb_invert: bool,
                 color_order: ColorOrder):
        pygame.init()
        self.screen = \
            pygame.display.set_mode((width * LED_SIZE,
                                     height * LED_SIZE))

        self._width, self._height = width, height
        self._buffer: List[List[Optional[Tuple[int, int, int]]]] = []
        for _ in range(height):
            self._buffer.append([None] * width)

    def start(self) -> None:
        pass

    def update(self, buffer: picographics.PicoGraphics) -> None:
        for y in range(self._height):
            for x in range(self._width):
                if buffer._buffer[y][x] != self._buffer[y][x]:
                    centre = (math.floor(x * LED_SIZE + LED_SIZE/2),
                              math.floor(y * LED_SIZE + LED_SIZE/2))
                    colour = buffer._buffer[y][x]
                    if colour is None:
                        colour = (0, 0, 0)
                    pygame.draw.circle(self.screen,
                                       colour,
                                       centre,
                                       LED_SIZE/2)

        pygame.display.flip()


class PanelType:
    pass


PANEL_GENERIC = PanelType()

COLOR_ORDER_RGB = ColorOrder()



#!/usr/bin/env micropython
# i75
# Copyright (C) 2023 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

try:
    ticks_diff = time.ticks_diff
    ticks_ms = time.ticks_ms
except AttributeError:
    def ticks_ms():
        return time.time_ns() / 1000000

    def ticks_diff(t1, t2):
        # TODO: Probably need to handle wrap around here.
        return t1 - t2


class DateTime:
    BASE: int = 0

    def __init__(self,
                 year: int,
                 month: int,
                 day: int,
                 weekday: int = 0,
                 hour: int = 0,
                 minute: int = 0,
                 second: int = 0,
                 microsecond: int = 0) -> None:
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DateTime):
            return False
        return self.year == other.year \
            and self.month == other.month \
            and self.day == other.day \
            and self.hour == other.hour \
            and self.minute == other.minute \
            and self.second == other.second

    def __str__(self):
        return f"{self.year}-{self.month:02n}-{self.day:02n} " \
             + f"{self.hour:02n}:{self.minute:02n}:{self.second:02n}"
